@@ -12,7 +12,6 @@ from enable.component_editor import ComponentEditor
 from pyface.image_resource import ImageResource
 from traitsui.api import (  # CloseAction,
     Action,
-    CheckListEditor,
     FileEditor,
     Group,
     HGroup,
@@ -21,6 +20,7 @@ from traitsui.api import (  # CloseAction,
     MenuBar,
     NoButtons,
     ObjectColumn,
+    RangeEditor,
     Separator,
     TableEditor,
     TextEditor,
@@ -62,13 +62,12 @@ traits_view = View(
                                 name="mod_type",
                                 label="Modulation",
                                 tooltip="line signalling/modulation scheme",
-                                editor=CheckListEditor(values=[(0, "NRZ"), (1, "Duo-binary"), (2, "PAM-4")]),
                             ),
                             Item(
                                 name="rlm",
                                 label="RLM",
                                 tooltip="relative level mismatch",
-                                enabled_when="mod_type[0] == 2"
+                                enabled_when="mod_type_ == 2"
                             ),
                             label="Rate && Modulation",
                             show_border=True,
@@ -387,9 +386,15 @@ traits_view = View(
                                 tooltip="Apply MLSD to recovered symbols, using Viterbi algorithm.",
                             ),
                             Item(
-                                name="rx_viterbi_symbols", label="# Symbols",
+                                name="rx_viterbi_symbols", label="Trellis Depth",
+                                enabled_when="rx_use_viterbi",
                                 tooltip="Number of symbols to include in MLSD trellis.",
                             ),
+                            # Item(
+                            #     name="rx_viterbi_fec", label="Use FEC",
+                            #     enabled_when="rx_use_viterbi",
+                            #     tooltip="Use FEC, as opposed to ISI, for Viterbi decoding.",
+                            # ),
                         ),
                         label="Native",
                         show_border=True,
@@ -734,8 +739,6 @@ traits_view = View(
                 ),
                 VGroup(  # Rx FFE
                     HGroup(
-                        Item(name="btn_disable_ffe", show_label=False, tooltip="Disable all FFE taps."),
-                        Item(name="btn_enable_ffe",  show_label=False, tooltip="Enable all FFE taps."),
                         Item(name="use_mmse", label="Use MMSE", tooltip="Use COM style MMSE optimization."),
                     ),
                     Item(
@@ -818,7 +821,7 @@ traits_view = View(
             ),
             Group(
                 Item("plots_H", editor=ComponentEditor(high_resolution=HIGH_RES), show_label=False),
-                label="Freq. Resp.",
+                label="Frequency",
                 id="plots_H",
             ),
             layout="tabbed",
@@ -846,6 +849,22 @@ traits_view = View(
                 label="Bathtubs",
                 id="plots_bathtub",
             ),
+            VGroup(
+                Item("plot_viterbi", editor=ComponentEditor(high_resolution=HIGH_RES), show_label=False,
+                     enabled_when="rx_use_viterbi == True and n_errs_viterbi != -1",),
+                HGroup(
+                    Item("trellis_err_select", label="Error",
+                         editor=RangeEditor(low=0, high_name="trellis_max_err", mode="xslider"),),
+                    Item(label="of"),
+                    Item("trellis_max_err", style="readonly", show_label=False),
+                    Item("trellis_pan_control", label="Position", springy=True,
+                         editor=RangeEditor(low=0, high_name="trellis_max_x", mode="xslider"),),
+                    enabled_when="rx_use_viterbi == True and n_errs_viterbi != -1",
+                ),
+                label="Viterbi",
+                id="plots_viterbi",
+                enabled_when="rx_use_viterbi == True",  # ToDo: Doesn't work. Alternative?
+            ),
             layout="tabbed",
             label="Results",
             id="results",
