@@ -1,7 +1,8 @@
 """
 General purpose utilities for PyBERT.
 
-Original author: David Banas <capn.freako@gmail.com>  
+Original author: David Banas <capn.freako@gmail.com>
+
 Original date:   June 16, 2024
 
 Copyright (c) 2024 David Banas; all rights reserved World wide.
@@ -13,9 +14,10 @@ A refactoring of the `pybert.utility` module, as per Issue #147.
 """
 
 from .channel import *
+from .functional import *
 from .ibisami import *
 from .jitter import *
-from .math import *
+from .math import *  # type: ignore
 from .python import *
 from .sigproc import *
 from .sparam import *
