@@ -9,6 +9,7 @@ Copyright (c) 2025 David Banas; all rights reserved World wide.
 
 from typing import Any, TypeVar
 
+S = TypeVar('S', Any, Any)
 T = TypeVar('T', Any, Any)
 
 
@@ -24,3 +25,17 @@ def fst(xs: tuple[T, ...]) -> T:
     """
 
     return xs[0]
+
+
+def snd(xs: tuple[S, T, ...]) -> T:  # type: ignore
+    """
+    Python translation of Haskell ``snd()`` function.
+
+    Args:
+        xs: Tuple having second element of type: ``T``.
+
+    Returns:
+        Second element of given tuple.
+    """
+
+    return xs[1]
